
#include "cuda_runtime.h"
#include "device_launch_parameters.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <string>

#define TILE 16

/***************************************************
 *@brief: Naive Matrix Matrix Multiplication in CUDA  
 **************************************************/
__global__ void MM_Mul_Default(int* a, int* b, int* c, int N)
{
    int tix = threadIdx.x + blockDim.x * blockIdx.x;
    int tiy = threadIdx.y + blockDim.y * blockIdx.y;


    if ((tix < N) && (tiy < N)) {
        int tid = tix + tiy * N;

        int sum = 0;
        for (int i = 0; i < N; i++) {
            int temp1 = a[tiy * N + i]; // Row
            int temp2 = b[N * i + tix]; // Col
            sum += temp1 * temp2;
        }

        c[tid] = sum;
    }

}

/******************************************************************************
 *@brief Optimized Matrix Matrix Multiplication by utilization of Shared Memory  
 *****************************************************************************/
__global__ void MM_MUl_Optimized(int* a, int* b, int* c, int N,int Phases) {

    // -----------------------------
    // ------- Grid Indexing -------
    unsigned int tx  = threadIdx.x;
    unsigned int ty  = threadIdx.y;
    unsigned int tix = tx + blockDim.x * blockIdx.x;
    unsigned int tiy = ty + blockDim.y * blockIdx.y;
    

    // ------------------------------------------------
    // ------ Statically Declared Shared Memory -------
    __shared__ float matrix_a[TILE][TILE];
    __shared__ float matrix_b[TILE][TILE];

    // ---------------------------------------
    // ------- Check Processing Offset -------
    if ((tix < N) & (tiy < N)) {
        unsigned int Lindex = tiy * N + tix;
        int sum = 0;

        // --------------------------
        // ------- Phase Loop -------
        for (int i = 0; i < Phases; i++) {

            // --------------------------------------------------
            // ------- Load Phase Data into Shared Memory ------- 
            matrix_a[ty][tx] = a[tiy * N             + i * TILE + tx];
            matrix_b[ty][tx] = b[(i * TILE + ty) * N + tix];

            __syncthreads();

            // -----------------------------------------
            // ------- Loop Across Shared Memory -------
            for (unsigned int j = 0; j < TILE; j++) { 
                int Temp1 = matrix_a[ty][j]; // Row
                int Temp2 = matrix_b[j][tx]; // Col
                sum +=  Temp1 * Temp2;
            }
            __syncthreads();
        }

        // ----------------------------------------
        // ------- Save Data Back to Global -------
        c[Lindex] = sum;
    }

}

/*************************
 * @brief: Device Pointers 
 *************************/
struct Dev_t{
    int* Mat_A;
    int* Mat_B;
    int* Mat_C;
};

/***********************
 * @brief: Host Pointers  
 **********************/
struct Host_t {
    int* Mat_A;
    int* Mat_B;
    int* Mat_C;
};

/********************************************
 * @Brief: Verifies MM Multiplication Results 
 ********************************************/
void CPU_MM_Verify(int* a, int* b, int* Ref,int N) {
    int NumErrs = 0;
    for (int y = 0; y < N; y++) {
        for (int x = 0; x < N; x++) {

            // -------------------------------------
            // ------- Compute Single Element ------
            int Res = 0;
            for (int i = 0; i < N; i++) {
                Res += a[y * N + i] * b[N * i + x];
            }

            // ----------------------
            // ------- Verify -------
            if (Res != Ref[y * N + x]) {
                NumErrs++;
            }
        }
    }

    if (NumErrs > 0) {
        std::cout << "Matrix MM Mult Verification Failed with " << std::to_string(NumErrs) << " Errors!" << std::endl;
    }else {
        std::cout << "Matrix MM Mult Verification Success" << std::endl;
    }

}




int main(int argc,char **argv){

    int Loops = 500;         // Loops Per Single Test
    int N = 1024;            // Matrix Size: [N*N]
    Dev_t Dev;
    Host_t Host;

    cudaEvent_t Start;
    cudaEvent_t Stop;

    float Default_Elapsed_Time;
    float Optimized_Elapsed_Time;
    cudaEventCreate(&Start);
    cudaEventCreate(&Stop);

    // ------------------------------------
    // -------- Allocate GPU Memory -------
    cudaMalloc((void**)&Dev.Mat_A, N * N * sizeof(int));
    cudaMalloc((void**)&Dev.Mat_B, N * N * sizeof(int));
    cudaMalloc((void**)&Dev.Mat_C, N * N * sizeof(int));

    // ----------------------------------
    // ------- Allocate CPU Memory ------
    cudaHostAlloc((void**)&Host.Mat_A, N * N * sizeof(int), cudaHostAllocDefault);
    cudaHostAlloc((void**)&Host.Mat_B, N * N * sizeof(int), cudaHostAllocDefault);
    cudaHostAlloc((void**)&Host.Mat_C, N * N * sizeof(int), cudaHostAllocDefault);

    // ----------------------------------
    // ------ Randomize Input Data ------
    srand(time(NULL));
    for (int i = 0; i < N; i++) {
        for (int j = 0; j < N; j++) {
            unsigned int Index = j + i * N;
            Host.Mat_A[Index] = rand() % 100;
            Host.Mat_B[Index] = rand() % 100;
        }
    }

    // -----------------------------------
    // ------- Copy Data To Device -------
    cudaMemcpy(Dev.Mat_A, Host.Mat_A, N * N * sizeof(int), cudaMemcpyHostToDevice);
    cudaMemcpy(Dev.Mat_B, Host.Mat_B, N * N * sizeof(int), cudaMemcpyHostToDevice);

    // ---------------------------------------
    // ------- Setup Kernel Dimensions -------
    dim3 Threads = dim3(TILE, TILE, 1);
    dim3 Blocks = dim3(N / TILE, N / TILE, 1);

    // -----------------------------------------------------
    // ------- ------- Launch Default Kernel ------- -------
    cudaEventRecord(Start);
    for (int i = 0; i < Loops; i++) {
        MM_Mul_Default << <Blocks, Threads >> > (Dev.Mat_A, Dev.Mat_B, Dev.Mat_C, N);
    }
    cudaEventRecord(Stop);
    cudaEventSynchronize(Stop);
    cudaEventElapsedTime(&Default_Elapsed_Time, Start, Stop);


    // -------------------------------------
    // ------ Copy Data back to Host -------
    cudaMemcpy(Host.Mat_C, Dev.Mat_C, N * N * sizeof(int), cudaMemcpyDeviceToHost);

    // ------------------------------
    // ------- Verify Results -------
    CPU_MM_Verify(Host.Mat_A , Host.Mat_B, Host.Mat_C, N);

    // --------------------------------------------------------
    // ------- ------- Launch Optimized Kernel ------- --------
    cudaEventRecord(Start);
    for (int i = 0; i < Loops; i++) {
        MM_MUl_Optimized << <Blocks, Threads >> > (Dev.Mat_A, Dev.Mat_B, Dev.Mat_C, N, N / TILE);
    }
    cudaEventRecord(Stop);
    cudaEventSynchronize(Stop);
    cudaEventElapsedTime(&Optimized_Elapsed_Time, Start, Stop);


    // -------------------------------------
    // ------ Copy Data back to Host -------
    cudaMemcpy(Host.Mat_C, Dev.Mat_C, N * N * sizeof(int), cudaMemcpyDeviceToHost);

    // ------------------------------
    // ------- Verify Results -------
    CPU_MM_Verify(Host.Mat_A, Host.Mat_B, Host.Mat_C, N);

    std::cout << "Default Time      : " << std::to_string(Default_Elapsed_Time / static_cast<float>(Loops)) << " ms" << std::endl;
    std::cout << "Optimized Time    : " << std::to_string(Optimized_Elapsed_Time / static_cast<float>(Loops)) << " ms" << std::endl;

    // ---------------------------------
    // ------- Cleanup Resources -------
    cudaEventDestroy(Start);
    cudaEventDestroy(Stop);
    cudaFreeHost(Host.Mat_A);
    cudaFreeHost(Host.Mat_B);
    cudaFreeHost(Host.Mat_C);
    cudaFree(Dev.Mat_A);
    cudaFree(Dev.Mat_B);
    cudaFree(Dev.Mat_C);



}

